/*****************************************************************************/

/*
 *      tests.c  --  Actual stress tests.
 *
 *      Copyright (C) 2000
 *        Thomas Sailer (sailer@ife.ee.ethz.ch)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  $Id: tests.c,v 1.7 2000/02/09 12:36:53 tom Exp $	 
 *
 */

/*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* AIX requires this to be the first thing in the file.  */
#ifndef __GNUC__
# if HAVE_ALLOCA_H
#  include <alloca.h>
# else
#  ifdef _AIX
#pragma alloca
#  else
#   ifndef alloca /* predefined by HP cc +Olibcalls */
char *alloca ();
#   endif
#  endif
# endif
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "usbdrv.h"
#include "usbtests.h"

/* ---------------------------------------------------------------------- */

#define USB_TYPE_STANDARD               (0x00 << 5)
#define USB_TYPE_CLASS                  (0x01 << 5)
#define USB_TYPE_VENDOR                 (0x02 << 5)
#define USB_TYPE_RESERVED               (0x03 << 5)

#define USB_RECIP_DEVICE                0x00
#define USB_RECIP_INTERFACE             0x01
#define USB_RECIP_ENDPOINT              0x02
#define USB_RECIP_OTHER                 0x03

#define USB_DIR_OUT                     0
#define USB_DIR_IN                      0x80

/* ---------------------------------------------------------------------- */

static int comparemem(const void *mem1, const void *mem2, size_t sz)
{
	unsigned int i, j;

	if (!memcmp(mem1, mem2, sz))
		return 0;
	fprintf(stderr, "Memory comparison error:\n");
	for (i = 0; i < sz; i += 8) {
		fprintf(stderr, "%04x: ", i);
		for (j = i; j < i+8; j++) {
			if (j >= sz) {
				fprintf(stderr, "   ");
				continue;
			}
			fprintf(stderr, " %02x", ((unsigned char *)mem1)[j]);
		}
		fprintf(stderr, "  -- ");
		for (j = i; j < i+8; j++) {
			if (j >= sz)
				break;
			fprintf(stderr, " %02x", ((unsigned char *)mem2)[j]);
		}
		fprintf(stderr, "\n");
	}
	return -1;
}

static int reset_bulk(struct usbdevice *dev, unsigned int ep)
{
	unsigned char buf[2];
	int r;

	r = usb_control_msg(dev, USB_RECIP_ENDPOINT|USB_TYPE_VENDOR|USB_DIR_IN, 0xe2, 0, ep, sizeof(buf), buf, 1000);
	if (r != 0) {
		fprintf(stderr, "control_msg(0xc2, 0xe2, %u, %u) failed with %d, errno %s (%d)\n",
			ep, sizeof(buf), r, strerror(errno), errno);
		return -1;
	}
	return 0;
}

static int display_status(struct usbdevice *dev)
{
	unsigned char buf[6];
	int r;

	r = usb_control_msg(dev, USB_RECIP_DEVICE|USB_TYPE_VENDOR|USB_DIR_IN, 0xe1, 0, 0, sizeof(buf), buf, 1000);
	if (r != sizeof(buf)) {
		fprintf(stderr, "control_msg(0xc0, 0xe1, %u) failed with %d, errno %s (%d)\n",
			sizeof(buf), r, strerror(errno), errno);
		return -1;
	}
	printf("device status: bulk blocks %d bulk2 blocks %d irqcount 0x%02x sofcount 0x%02x control length %d leddiv 0x%02x\n",
	       buf[0], buf[1], buf[2], buf[3], buf[4], buf[5]);
	return 0;
}

/* ---------------------------------------------------------------------- */

static int reset(struct usbdevice *dev, char *param)
{
	return anchor_reset(dev);
}

/* ---------------------------------------------------------------------- */

static int controlecho(struct usbdevice *dev, char *param)
{
	static const char spinchr[] = "/-\\|";
	unsigned char buf[48];
	unsigned char bufr[sizeof(buf)];
	unsigned long seq = 0;
	int r;

	printf("Control Echo test\n");
	for (;;) {
		for (r = 0; r < sizeof(buf); r++)
			buf[r] = random();
		r = usb_control_msg(dev, USB_RECIP_DEVICE|USB_TYPE_VENDOR|USB_DIR_OUT, 0xe0, 0, 0, sizeof(buf), buf, 1000);
		if (r != sizeof(buf)) {
			fprintf(stderr, "control_msg(0x40, 0xe0, %u) failed with %d, errno %s (%d) sequence %u\n",
				sizeof(buf), r, strerror(errno), errno, seq);
			return -1;
		}
		r = usb_control_msg(dev, USB_RECIP_DEVICE|USB_TYPE_VENDOR|USB_DIR_IN, 0xe0, 0, 0, sizeof(bufr), bufr, 1000);
		if (r != sizeof(buf)) {
			fprintf(stderr, "control_msg(0xc0, 0xe0, %u) failed with %d, errno %s (%d) sequence %u\n",
				sizeof(bufr), r, strerror(errno), errno, seq);
			return -1;
		}
		if (comparemem(buf, bufr, sizeof(buf))) {
			fprintf(stderr, "Data error at sequence %u\n", seq);
			return -1;
		}
		if (!(seq & 63)) {
			printf("%5u %c\r", seq, spinchr[(seq >> 6) & 3]);
			fflush(stdout);
		}
		seq++;
	}
}

/* ---------------------------------------------------------------------- */

static int bulkecho(struct usbdevice *dev, char *param)
{
	static const char spinchr[] = "/-\\|";
	unsigned char buf[1024];
	unsigned char bufr[1024];
	unsigned long seq = 0;
	unsigned int size = 480, intf = 1;
	char *e, *p;
	int r;

	if (param) {
		p = strtok(param, ",");
		while (p) {
			if (!strncmp(p, "if=", 3)) {
				intf = strtoul(p+3, &e, 0);
				if (!e || e[0] || intf < 1 || intf > 3)
					goto paramerror;
			} else if (!strncmp(p, "sz=", 3)) {
				size = strtoul(p+3, &e, 0);
				if (!e || e[0] || size < 1 || size > sizeof(buf))
					goto paramerror;
			} else
				goto paramerror;
			p = strtok(NULL, ",");
		}
	}
	printf("Bulk Echo test, packet size %u\n", size);
	if (usb_claiminterface(dev, intf)) {
		perror("usb_claiminterface");
		return -1;
	}
	if (usb_setinterface(dev, intf, 0)) {
		perror("usb_setinterface");
		return -1;
	}
	if (display_status(dev))
		return -1;
	if (display_status(dev))
		return -1;
	if (reset_bulk(dev, USB_DIR_OUT|(2*intf)))
		return -1;
	if (display_status(dev))
		return -1;
	for (;;) {
		for (r = 0; r < size; r++)
			buf[r] = random();
		r = usb_bulk_msg(dev, USB_DIR_OUT|(2*intf), size, buf, 1000);
		if (r != size) {
			fprintf(stderr, "usb_bulk_msg(0x%02x, %u) failed with %d, errno %s (%d) sequence %u\n",
				USB_DIR_OUT|(2*intf), size, r, strerror(errno), errno, seq);
			return -1;
		}
		r = usb_bulk_msg(dev, USB_DIR_IN|(2*intf), sizeof(bufr), bufr, 1000);
		if (r != size) {
			fprintf(stderr, "usb_bulk_msg(0x%02x, %u) failed with %d, errno %s (%d) sequence %u\n",
				USB_DIR_IN|(2*intf), sizeof(bufr), r, strerror(errno), errno, seq);
			return -1;
		}
		if (comparemem(buf, bufr, size)) {
			fprintf(stderr, "Data error at sequence %u\n", seq);
			return -1;
		}
		if (!(seq & 63)) {
			printf("%5u %c\r", seq, spinchr[(seq >> 6) & 3]);
			fflush(stdout);
		}
		seq++;
	}
	return 0;

  paramerror:
	fprintf(stderr, "parameter error, usage eg: if=1,sz=32\n");
	return -1;
}

/* ---------------------------------------------------------------------- */

static int abulkecho(struct usbdevice *dev, char *param)
{
	static const char spinchr[] = "/-\\|";
	unsigned char buf[1024];
	unsigned char bufr[1024];
	unsigned long seq = 0, seqr = 0, seqrx;
	unsigned int size = 480, intf = 1;
	struct usbdevfs_urb txurb, rxurb, *purb;
	char *e, *p;
	int r;

	if (param) {
		p = strtok(param, ",");
		while (p) {
			if (!strncmp(p, "if=", 3)) {
				intf = strtoul(p+3, &e, 0);
				if (!e || e[0] || intf < 1 || intf > 3)
					goto paramerror;
			} else if (!strncmp(p, "sz=", 3)) {
				size = strtoul(p+3, &e, 0);
				if (!e || e[0] || size < 1 || size > sizeof(buf))
					goto paramerror;
			} else
				goto paramerror;
			p = strtok(NULL, ",");
		}
	}
	printf("Asynchronous Bulk Echo test, packet size %u\n", size);
	if (usb_claiminterface(dev, intf)) {
		perror("usb_claiminterface");
		return -1;
	}
	if (usb_setinterface(dev, intf, 0)) {
		perror("usb_setinterface");
		return -1;
	}
	if (display_status(dev))
		return -1;
	if (display_status(dev))
		return -1;
	if (reset_bulk(dev, USB_DIR_OUT|(2*intf)))
		return -1;
	if (display_status(dev))
		return -1;
	memset(&rxurb, 0, sizeof(rxurb));
	rxurb.type = USBDEVFS_URB_TYPE_BULK;
	rxurb.endpoint = USB_DIR_IN|(2*intf);
	rxurb.flags = 0;
	rxurb.buffer = bufr;
	rxurb.buffer_length = sizeof(bufr);
	rxurb.signr = 0;
	rxurb.usercontext = NULL;
	memset(&txurb, 0, sizeof(txurb));
	txurb.type = USBDEVFS_URB_TYPE_BULK;
	txurb.endpoint = USB_DIR_OUT|(2*intf);;
	txurb.flags = 0;
	txurb.buffer = buf;
	txurb.buffer_length = size;
	txurb.signr = 0;
	txurb.usercontext = NULL;
	for (r = 0; r < size; r++)
		buf[r] = random();
	buf[0] = seq;
	buf[1] = seq >> 8;
	buf[2] = seq >> 16;
	buf[3] = seq >> 24;
	seq++;
	if (usb_submiturb(dev, &rxurb)) {
		perror("usb_submiturb");
		return -1;
	}
	if (usb_submiturb(dev, &txurb)) {
		perror("usb_submiturb");
		goto err;
	}
	for (;;) {
		if (!(purb = usb_reapurb(dev, 0))) {
			perror("usb_reapurb");
			goto err;
		}
		if (purb == &rxurb) {
			seqrx = bufr[0] | (bufr[1] << 8) | (bufr[2] << 16) | (bufr[3] << 24);
			if (rxurb.status || rxurb.actual_length != size) {
				fprintf(stderr, "Receive error: status %d actlen %d rxseq 0x%08x expected seq 0x%08x\n",
					rxurb.status, rxurb.actual_length, seqrx, seqr);
				seqr++;
			} else {
				if (seqrx != seqr)
					fprintf(stderr, "RX sequence error, received 0x%08x expected 0x%08x\n", seqrx, seqr);
				seqr = seqrx + 1;
			}
			if (usb_submiturb(dev, &rxurb)) {
				perror("usb_submiturb");
				goto err;
			}
		} else if (purb == &txurb) {
			if (txurb.status)
				fprintf(stderr, "Transmit error: status %d actlen %d seq 0x%08x\n",
					txurb.status, txurb.actual_length, seq);
			buf[0] = seq;
			buf[1] = seq >> 8;
			buf[2] = seq >> 16;
			buf[3] = seq >> 24;
			seq++;
			if (usb_submiturb(dev, &txurb)) {
				perror("usb_submiturb");
				goto err;
			}
			if (!(seq & 63)) {
				printf("%8u %8u %c\r", seq, seqr, spinchr[(seq >> 6) & 3]);
				fflush(stdout);
			}
		} else {
			fprintf(stderr, "reapurb: returned invalid URB! ret %p rx %p tx %p\n", purb, &rxurb, &txurb);
		}
	}
	return 0;

  paramerror:
	fprintf(stderr, "parameter error, usage eg: if=1,sz=32\n");
	return -1;

  err:
	if (usb_discardurb(dev, &rxurb))
		perror("usb_discardurb");
	if (usb_discardurb(dev, &txurb))
		perror("usb_discardurb");
	return -1;
}

/* ---------------------------------------------------------------------- */

#define MAXISOFRAMES 64

static int isoecho(struct usbdevice *dev, char *param)
{
	static const char spinchr[] = "/-\\|";
	unsigned char buf[1024];
	unsigned char bufr[1024];
	unsigned long seq = 0, seqr = 0, seqrx;
	unsigned int size = 480, intf = 4, frames = 16, maxsize, f, u;
	struct usbdevfs_urb *purb;
	struct {
		struct usbdevfs_urb urb;
		struct usbdevfs_iso_packet_desc iso[MAXISOFRAMES];
	} txurb[2], rxurb[2];
	struct usbdevfs_iso_packet_desc *iso;
	char *e, *p;
	unsigned char *bp;
	int r;

	if (param) {
		p = strtok(param, ",");
		while (p) {
			if (!strncmp(p, "if=", 3)) {
				intf = strtoul(p+3, &e, 0);
				if (!e || e[0] || intf < 4 || intf > 5)
					goto paramerror;
			} else if (!strncmp(p, "sz=", 3)) {
				size = strtoul(p+3, &e, 0);
				if (!e || e[0] || size < 4)
					goto paramerror;
			} else if (!strncmp(p, "fr=", 3)) {
				frames = strtoul(p+3, &e, 0);
				if (!e || e[0] || frames < 1)
					goto paramerror;
			} else
				goto paramerror;
			p = strtok(NULL, ",");
		}
	}
	maxsize = (intf == 4) ? 496 : 16;
	if (size > maxsize || frames > MAXISOFRAMES)
		goto paramerror;
	printf("Isochronous Echo test, packet size %u, number of frames per URB %u\n", size, frames);
	if (usb_claiminterface(dev, intf)) {
		perror("usb_claiminterface");
		return -1;
	}
	if (usb_setinterface(dev, intf, 1)) {
		perror("usb_setinterface");
		return -1;
	}
#if 0
	if (display_status(dev))
		return -1;
#endif
	memset(&rxurb[0], 0, sizeof(rxurb[0]));
	rxurb[0].urb.type = USBDEVFS_URB_TYPE_ISO;
	rxurb[0].urb.endpoint = USB_DIR_IN|(intf+4);
	rxurb[0].urb.flags = USBDEVFS_URB_ISO_ASAP;
	rxurb[0].urb.buffer_length = maxsize * frames;  
	rxurb[0].urb.number_of_packets = frames;
	rxurb[0].urb.signr = 0;
	rxurb[0].urb.usercontext = NULL;
	for (f = 0; f < frames; f++)
		rxurb[0].urb.iso_frame_desc[f].length = maxsize;
	rxurb[1] = rxurb[0];
	rxurb[0].urb.buffer = alloca(rxurb[0].urb.buffer_length);
	rxurb[1].urb.buffer = alloca(rxurb[1].urb.buffer_length);
	memset(&txurb[0], 0, sizeof(txurb[0]));
	txurb[0].urb.type = USBDEVFS_URB_TYPE_ISO;
	txurb[0].urb.endpoint = USB_DIR_OUT|(intf+4);;
	txurb[0].urb.flags = USBDEVFS_URB_ISO_ASAP;
	txurb[0].urb.buffer_length = size * frames;
	txurb[0].urb.number_of_packets = frames;
	txurb[0].urb.signr = 0;
	txurb[0].urb.usercontext = NULL;
	for (f = 0; f < frames; f++)
		txurb[0].urb.iso_frame_desc[f].length = size;
	txurb[1] = txurb[0];
	txurb[0].urb.buffer = alloca(txurb[0].urb.buffer_length);
	txurb[1].urb.buffer = alloca(txurb[1].urb.buffer_length);
	for (u = 0; u < 2; u++) {
		bp = txurb[u].urb.buffer;
		for (f = 0; f < frames; f++, bp += size) {
			for (r = 0; r < size; r++)
				bp[r] = random();
			bp[0] = seq;
			bp[1] = seq >> 8;
			bp[2] = seq >> 16;
			bp[3] = seq >> 24;
			seq++;
		}
	}
#if 1
	if (usb_submiturb(dev, &rxurb[0].urb)) {
		perror("usb_submiturb");
		return -1;
	}
	if (usb_submiturb(dev, &rxurb[1].urb)) {
		perror("usb_submiturb");
		goto err;
	}
#endif
	if (usb_submiturb(dev, &txurb[0].urb)) {
		perror("usb_submiturb");
		goto err;
	}
	if (usb_submiturb(dev, &txurb[1].urb)) {
		perror("usb_submiturb");
		goto err;
	}
	for (;;) {
		if (!(purb = usb_reapurb(dev, 0))) {
			perror("usb_reapurb");
			goto err;
		}
		if (purb == &rxurb[0].urb || purb == &rxurb[1].urb) {
			if (purb->status || purb->error_count)
				fprintf(stderr, "Receive error: status %d count: %d\n", purb->status, purb->error_count);
			bp = purb->buffer;
			iso = purb->iso_frame_desc;
			for (f = 0; f < frames; f++, bp += maxsize, iso++) {
				seqrx = bp[0] | (bp[1] << 8) | (bp[2] << 16) | (bp[3] << 24);
				if (iso->status || iso->actual_length != size) {
					fprintf(stderr, "Receive error: frame %d status %d actlen %d rxseq 0x%08x expected seq 0x%08x\n",
						f, iso->status, iso->actual_length, seqrx, seqr);
					seqr++;
					continue;
				}
				if (seqrx != seqr)
					fprintf(stderr, "Receive sequence error in frame %d, received 0x%08x expected 0x%08x actlen %d len %d\n", f, seqrx, seqr, iso->actual_length, iso->length);
				seqr = seqrx + 1;
			}
			if (usb_submiturb(dev, purb)) {
				perror("usb_submiturb");
				goto err;
			}
		} else if (purb == &txurb[0].urb || purb == &txurb[1].urb) {
			if (purb->status || purb->error_count)
				fprintf(stderr, "Transmit error: status %d count %d seq 0x%08x\n",
					purb->status, purb->error_count, seq);
			bp = purb->buffer;
			iso = purb->iso_frame_desc;
			for (f = 0; f < frames; f++, bp += size, iso++) {
				if (iso->status || iso->actual_length != size)
					fprintf(stderr, "Transmit error: frame %d status %d actlen %d\n",
						f, iso->status, iso->actual_length);
				bp[0] = seq;
				bp[1] = seq >> 8;
				bp[2] = seq >> 16;
				bp[3] = seq >> 24;
				seq++;
			}
			if (usb_submiturb(dev, purb)) {
				perror("usb_submiturb");
				goto err;
			}
			if (!(seq & 255)) {
				printf("%8u %8u %c\r", seq, seqr, spinchr[(seq >> 8) & 3]);
				fflush(stdout);
			}
		} else {
			fprintf(stderr, "reapurb: returned invalid URB! ret %p rx0 %p rx1 %p tx0 %p tx1 %p\n",
				purb, &rxurb[0], &rxurb[1], &txurb[0], &txurb[1]);
		}
	}
	return 0;

  paramerror:
	fprintf(stderr, "parameter error, usage eg: if=4,sz=32,fr=12\n");
	return -1;

  err:
	if (usb_discardurb(dev, &rxurb[0].urb))
		perror("usb_discardurb");
	if (usb_discardurb(dev, &rxurb[1].urb))
		perror("usb_discardurb");
	if (usb_discardurb(dev, &txurb[0].urb))
		perror("usb_discardurb");
	if (usb_discardurb(dev, &txurb[1].urb))
		perror("usb_discardurb");
	return -1;
}

/* ---------------------------------------------------------------------- */

const struct testtab testtab[] = {
	{ "controlecho", "Echo using synchronous control messages", controlecho },
	{ "c", "Same as \"controlecho\"", controlecho },
	{ "bulkecho", "Echo using synchronous bulk messages", bulkecho },
	{ "b", "Same as \"bulkecho\"", bulkecho },
	{ "abulkecho", "Echo using asynchronous bulk messages", abulkecho },
	{ "ab", "Same as \"abulkecho\"", abulkecho },
	{ "isoecho", "Echo using isochronous messages", isoecho },
	{ "i", "Same as \"isoecho\"", isoecho },
	{ "reset", "Reset AnchorChips device to default state", reset },
	{ "r", "Same as \"reset\"", reset },
	{ NULL, NULL, NULL }
};
