/*****************************************************************************/

/*
 *      usbstress.c  --  USB stress test utility.
 *
 *      Copyright (C) 2000
 *        Thomas Sailer (sailer@ife.ee.ethz.ch)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  $Id: usbstress.c,v 1.3 2000/01/11 11:07:52 tom Exp $	 
 *
 */

/*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdarg.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#include "usbdrv.h"
#include "usbtests.h"

/* ---------------------------------------------------------------------- */

static unsigned verboselevel = 0;
static int quit = 0;

/* ---------------------------------------------------------------------- */

int lprintf(unsigned vl, const char *format, ...)
{
        va_list ap;
        int r;

        if (vl > verboselevel)
                return 0;
	va_start(ap, format);
#ifdef HAVE_VSYSLOG
        if (syslogmsg) {
		static const int logprio[] = { LOG_ERR, LOG_INFO };
		vsyslog((vl > 1) ? LOG_DEBUG : logprio[vl], format, ap);
		r = 0;
	} else
#endif
                r = vfprintf(stderr, format, ap);
	va_end(ap);
        return r;
}

int tprintf(const char *format, ...)
{
	va_list ap;
	int i;

        va_start(ap, format);
        i = vprintf(format, ap);
        va_end(ap);
	fflush(stdout);
	return i;
}

int idle_callback(unsigned us_delay)
{
	if (us_delay)
		usleep(us_delay);
	return quit;
}

static void signal_quit(int signum)
{
	quit = 1;
exit(0); /* for now */
}

/* ---------------------------------------------------------------------- */

static void signal_disconn(int signr, siginfo_t *info, void *ucontext)
{
	if (!info) {
		printf("Device disconnected, but no siginfo\n");
		exit(1);
	}
	printf("Device disconnected, si_signo %d si_errno %d si_code %d context %p\n",
	       info->si_signo, info->si_errno, info->si_code, info->si_addr);
	exit(1);
}

/* ---------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
        static const struct option long_options[] = {
                { "usbdevfs", 1, 0, 'D' },
		{ 0, 0, 0, 0 }
        };
	int c, err = 0, busnum = 1, devnum = -1, fdownload = 0;
	const struct testtab *tst = NULL;
	char *par = NULL;
	struct usbdevice *dev;
	struct sigaction sigact;
	struct usb_device_descriptor desc;

	printf("usbstress V" VERSION " (C) 2000 by Thomas Sailer\n");
	while ((c = getopt_long(argc, argv, "D:d:b:fvp:t:", long_options, NULL)) != EOF) {
		switch (c) {
		case 'D':
			usb_devicefs_mountpoint = optarg;
			break;

		case 'v':
			verboselevel++;
			break;

		case 'p':
			par = optarg;
			break;

		case 'd':
			devnum = strtoul(optarg, NULL, 0);
			break;

		case 'b':
			busnum = strtoul(optarg, NULL, 0);
			break;

		case 'f':
			fdownload = 1;
			break;

		case 't':
			if (tst != NULL) {
				err++;
				break;
			}
			for (tst = testtab; tst->name && strcmp(optarg, tst->name); tst++);
			if (!tst->name)
				err++;
			break;

		default:
			err++;
			break;
		}
	}
	if (err) {
		lprintf(0, "usage: %s [-v] [-D <drvpath>] [-p <testparams>] [-t <test>]\n"
			"  test options:\n", argv[0]);
		for (tst = testtab; tst->name; tst++)
			lprintf(0, "    %-15s%s\n", tst->name, tst->help);
		exit(1);
	}
	sigact.sa_flags = SA_SIGINFO;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_sigaction = signal_disconn;
	if (sigaction(SIGRTMIN, &sigact, NULL)) {
		perror("sigaction");
		exit(1);
	}
	/* first try to find an anchorchips device */
	if (!fdownload && tst) {
		if (devnum != -1)
			dev = usb_open_bynumber(busnum, devnum, 0x1234, 0x5678);
		else
			dev = usb_open(0x1234, 0x5678, 0);
		if (dev) {
#ifdef SIGHUP
			signal(SIGHUP, signal_quit);
#endif
			signal(SIGINT, signal_quit);
#ifdef SIGQUIT
			signal(SIGQUIT, signal_quit);
#endif
			signal(SIGTERM, signal_quit);
			if (usb_discsignal(dev, SIGRTMIN, (void *)0x12345678))
				perror("usb_discsignal");
			if (usb_getdevicedescriptor(dev, &desc))
				perror("usb_getdevicedescriptor");
			else
				usb_show_device_descriptor(stdout, &desc);
			err = tst->proc(dev, par);
			usleep(500000);
			exit(err ? 1 : 0);
		}
		fprintf(stderr, "No AnchorChips device with loaded firmware found, trying to download\n");
	}
	if (devnum != -1)
		dev = usb_open_bynumber(busnum, devnum, 0x0547, 0x2131);
	else
		dev = usb_open(0x0547, 0x2131, 0);
	if (!dev) {
		fprintf(stderr, "No \"virgin\" AnchorChips device found\n");
		exit(1);
	}
	err = anchor_download(dev);
	exit(err ? 1 : 0);
}
